import pytest
import tempfile
from tonbo import DbOption, Column, DataType, Record, TonboDB, Bound
from tonbo.error import RepeatedCommitError, WriteConflictError


@Record
class User:
    age = Column(DataType.Int8, name="age", primary_key=True)
    height = Column(DataType.Int16, name="height", nullable=True)
    weight = Column(DataType.Int32, name="weight", nullable=False)


def build_db():
    temp_dir = tempfile.TemporaryDirectory()
    return TonboDB(DbOption(temp_dir.name), User())


@pytest.mark.asyncio
async def test_txn_write():
    db = build_db()
    txn = await db.transaction()
    for i in range(0, 100):
        txn.insert(User(age=i, height=i * 10, weight=i * 20))
    await txn.commit()


@pytest.mark.asyncio
async def test_txn_read_write():
    db = build_db()
    txn = await db.transaction()
    for i in range(0, 100):
        txn.insert(User(age=i, height=i * 10, weight=i * 20))
    for i in range(0, 100):
        txn.insert(User(age=i, height=i * 11, weight=i * 22))

    for i in range(0, 100):
        user = await txn.get(i)
        assert user == {"age": i, "height": i * 11, "weight": i * 22}


@pytest.mark.asyncio
async def test_txn_remove():
    db = build_db()
    txn = await db.transaction()
    for i in range(0, 100):
        if i % 2 == 0:
            txn.insert(User(age=i, height=i * 10, weight=i * 20))
        else:
            txn.remove(i)

    for i in range(0, 100):
        user = await txn.get(i)
        if i % 2 == 0:
            assert user == {"age": i, "height": i * 10, "weight": i * 20}
        else:
            assert user is None


@pytest.mark.asyncio
async def test_txn_scan():
    db = build_db()
    txn = await db.transaction()
    for i in range(0, 100):
        txn.insert(User(age=i, height=i * 10, weight=i * 20))

    scan = await txn.scan(Bound.Included(10), Bound.Excluded(75))
    i = 10
    async for user in scan:
        assert user == {"age": i, "height": i * 10, "weight": i * 20}
        i += 1
    assert i == 75


@pytest.mark.asyncio
async def test_txn_projection():
    db = build_db()
    txn = await db.transaction()
    for i in range(0, 100):
        txn.insert(User(age=i, height=i * 10, weight=i * 20))

    scan = await txn.scan(None, Bound.Excluded(75), projection=["age", "height"])
    i = 0
    async for user in scan:
        assert user == {"age": i, "height": i * 10, "weight": None}
        i += 1


@pytest.mark.asyncio
async def test_txn_limit():
    db = build_db()
    txn = await db.transaction()
    for i in range(0, 100):
        txn.insert(User(age=i, height=i * 10, weight=i * 20))

    scan = await txn.scan(None, Bound.Excluded(75), limit=10)
    i = 0
    async for user in scan:
        assert user == {"age": i, "height": i * 10, "weight": i * 20}
        i += 1
    assert i == 10


@pytest.mark.asyncio
async def test_repeated_commit():
    db = build_db()
    txn = await db.transaction()
    await txn.commit()

    with pytest.raises(RepeatedCommitError):
        await txn.commit()


@pytest.mark.asyncio
async def test_txn_write_conflict():
    db = build_db()
    txn = await db.transaction()
    txn2 = await db.transaction()
    for i in range(0, 10):
        txn.insert(User(age=i, height=i * 10, weight=i * 20))

    for i in range(0, 10):
        txn2.insert(User(age=i, height=i * 10, weight=i * 20))

    await txn2.commit()
    with pytest.raises(WriteConflictError):
        await txn.commit()


@pytest.mark.asyncio
async def test_txn_cannot_read_future():
    db = build_db()
    txn = await db.transaction()
    txn2 = await db.transaction()
    for i in range(0, 10):
        txn.insert(User(age=i, height=i * 10, weight=i * 20))

    for i in range(0, 10):
        txn2.insert(User(age=i, height=i * 20, weight=i * 40))

    await txn2.commit()
    for i in range(0, 10):
        user = await txn.get(i)
        # can not read data in the future
        assert user == { "age": i, "height": i * 10, "weight": i * 20 }

    txn3 = await db.transaction()
    for i in range(0, 10):
        user = await txn3.get(i)
        assert user == { "age": i, "height": i * 20, "weight": i * 40 }



import pytest
import tempfile
from tonbo import DbOption, Column, DataType, Record, RecordBatch, TonboDB


@Record
class User:
    age = Column(DataType.Int8, name="age", primary_key=True)
    height = Column(DataType.Int16, name="height", nullable=True)
    weight = Column(DataType.Int32, name="weight", nullable=False)


def build_db():
    temp_dir = tempfile.TemporaryDirectory()
    return TonboDB(DbOption(temp_dir.name), User())


@pytest.mark.asyncio
async def test_db_write():
    db = build_db()
    for i in range(0, 100):
        await db.insert(User(age=i, height=i * 10, weight=i * 20))


@pytest.mark.asyncio
async def test_db_read():
    db = build_db()
    for i in range(0, 100):
        await db.insert(User(age=i, height=i * 10, weight=i * 20))

    for i in range(0, 100):
        user = await db.get(i)
        assert user == {"age": i, "height": i * 10, "weight": i * 20}


@pytest.mark.asyncio
async def test_db_write_batch():
    db = build_db()
    batch = RecordBatch()
    for i in range(0, 100):
        batch.append(User(age=i, height=i * 10, weight=i * 20))
    await db.insert_batch(batch)

    for i in range(0, 100):
        user = await db.get(i)
        assert user == {"age": i, "height": i * 10, "weight": i * 20}


@pytest.mark.asyncio
async def test_db_remove():
    db = build_db()
    for i in range(0, 100):
        if i % 2 == 0:
            await db.insert(User(age=i, height=i * 10, weight=i * 20))
        else:
            await db.remove(i)

    for i in range(0, 100):
        user = await db.get(i)
        if i % 2 == 0:
            assert user == {"age": i, "height": i * 10, "weight": i * 20}
        else:
            assert user is None


@pytest.mark.asyncio
async def test_db_recover():
    temp_dir = tempfile.TemporaryDirectory()
    db = TonboDB(DbOption(temp_dir.name), User())
    for i in range(0, 100):
        await db.insert(User(age=i, height=i * 10, weight=i * 20))

    for i in range(0, 100):
        user = await db.get(i)
        assert user == {"age": i, "height": i * 10, "weight": i * 20}

    await db.flush_wal()

    db = TonboDB(DbOption(temp_dir.name), User())
    for i in range(0, 100):
        user = await db.get(i)
        assert user == {"age": i, "height": i * 10, "weight": i * 20}



import tempfile
import pytest
from tonbo import TonboDB, DbOption, Column, DataType, Record
from tonbo.fs import from_filesystem_path, parse, from_url_path, FsOptions
from tonbo.error import PathParseError


@Record
class User:
    id = Column(DataType.Int64, name="id", primary_key=True)
    name = Column(DataType.String, name="name")
    email = Column(DataType.String, name="email", nullable=True)
    age = Column(DataType.Int8, name="age")
    data = Column(DataType.Bytes, name="data")


def test_parse():
    assert parse("/foo/bar") == "foo/bar"
    with pytest.raises(PathParseError):
        parse("//foo/bar")


def test_from_url_path():
    assert from_url_path("foo%20bar") == "foo bar"
    assert from_url_path("foo%2F%252E%252E%2Fbar") == "foo/%2E%2E/bar"
    assert from_url_path("foo/%252E%252E/bar") == "foo/%2E%2E/bar"
    assert from_url_path("%48%45%4C%4C%4F") == "HELLO"
    with pytest.raises(PathParseError):
        from_url_path("foo/%2E%2E/bar")


@pytest.mark.asyncio
async def test_table_level_local():
    temp_dir = tempfile.TemporaryDirectory()
    temp_dir0 = tempfile.TemporaryDirectory()
    temp_dir1 = tempfile.TemporaryDirectory()

    option = DbOption(from_filesystem_path(temp_dir.name))
    option.level_path(0, from_filesystem_path(temp_dir0.name), FsOptions.Local())
    option.level_path(1, from_filesystem_path(temp_dir1.name), FsOptions.Local())

    option.immutable_chunk_num = 1
    option.major_threshold_with_sst_size = 3
    option.level_sst_magnification = 1
    option.max_sst_file_size = 1 * 1024

    db = TonboDB(option, User())
    for i in range(0, 1000):
        if i % 50 == 0:
            await db.flush()
        await db.insert(
            User(
                id=i,
                age=i % 128,
                name=str(i * 10),
                email=str(i * 20),
                data=b"Hello Tonbo!",
            )
        )

    for i in range(0, 1000):
        user = await db.get(i)
        assert user == {
            "id": i,
            "name": str(i * 10),
            "email": str(i * 20),
            "age": i % 128,
            "data": b"Hello Tonbo!",
        }



import os
import tempfile
import duckdb
import sqlite3
import pytest

from conftest import gen_string, gen_int, gen_bytes
from tonbo import Record, Column, DataType, TonboDB, DbOption
from tonbo.fs import from_filesystem_path

WRITE_TIME = 500000


@Record
class User:
    id = Column(DataType.Int64, name="id", primary_key=True)
    name = Column(DataType.String, name="name")
    email = Column(DataType.String, name="email", nullable=True)
    age = Column(DataType.UInt16, name="age")
    data = Column(DataType.Bytes, name="data")


def duckdb_write(auto_commit: bool):
    con = duckdb.connect()
    con.sql(
        "CREATE TABLE user (id INTEGER, name VARCHAR(20), email VARCHAR(20), age INTEGER, data VARCHAR(200))"
    )
    if not auto_commit:
        con.begin()
    for i in range(0, WRITE_TIME):
        con.execute(
            "INSERT INTO user VALUES (?, ?, ?, ?, ?)",
            [i, gen_string(20), gen_string(20), gen_int(0, 0xffff), gen_bytes(200)],
        )
    if not auto_commit:
        con.commit()
    con.close()


async def tonbo_write(auto_commit: bool):
    temp_dir = tempfile.TemporaryDirectory()

    option = DbOption(from_filesystem_path(temp_dir.name))

    db = TonboDB(option, User())
    if auto_commit:
        for i in range(0, WRITE_TIME):
            await db.insert(User(
                id=i,
                age=gen_int(0, 0xffff),
                name=gen_string(20),
                email=gen_string(20),
                data=gen_bytes(200),
            ))
    else:
        txn = await db.transaction()
        for i in range(0, WRITE_TIME):
            txn.insert(User(
                id=i,
                age=gen_int(0, 0xffff),
                name=gen_string(20),
                email=gen_string(20),
                data=gen_bytes(200),
            ))
        await txn.commit()

    await db.flush_wal()


def sqlite_write(auto_commit: bool):
    file = tempfile.NamedTemporaryFile()
    con = sqlite3.connect(file.name, autocommit=auto_commit)
    con.execute(
        "CREATE TABLE user (id INTEGER, name VARCHAR(20), email VARCHAR(20), age INTEGER, data VARCHAR(200))"
    )
    for i in range(0, WRITE_TIME):
        con.execute(
            "INSERT INTO user VALUES (?, ?, ?, ?, ?)",
            [i, gen_string(20), gen_string(20), gen_int(0, 0xffff), gen_bytes(200)],
        )
    con.commit()
    con.close()


@pytest.mark.parametrize("auto_commit", [True])
@pytest.mark.benchmark(group="autocommit")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_duckdb_autocommit(benchmark, auto_commit):
    benchmark(duckdb_write, auto_commit)


@pytest.mark.parametrize("auto_commit", [False])
@pytest.mark.benchmark(group="txn")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_duckdb(benchmark, auto_commit):
    benchmark(duckdb_write, auto_commit)


@pytest.mark.parametrize("auto_commit", [False])
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
@pytest.mark.benchmark(group="txn")
def test_tonbo(aio_benchmark, auto_commit):
    aio_benchmark(tonbo_write, auto_commit)


@pytest.mark.parametrize("auto_commit", [True])
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
@pytest.mark.benchmark(group="autocommit")
def test_tonbo_no_txn(aio_benchmark, auto_commit):
    aio_benchmark(tonbo_write, auto_commit)


@pytest.mark.parametrize("auto_commit", [True])
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
@pytest.mark.benchmark(group="autocommit")
def test_sqlite_autocommit(benchmark, auto_commit):
    benchmark(sqlite_write, auto_commit)


@pytest.mark.parametrize("auto_commit", [False])
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
@pytest.mark.benchmark(group="txn")
def test_sqlite(benchmark, auto_commit):
    benchmark(sqlite_write, auto_commit)



import pytest
import tempfile
from tonbo import DbOption, Column, DataType, Record, TonboDB


@Record
class User:
    id = Column(DataType.Int64, name="id", primary_key=True)
    name = Column(DataType.String, name="name")
    email = Column(DataType.String, name="email", nullable=True)
    age = Column(DataType.Int8, name="age")
    data = Column(DataType.Bytes, name="data")


@pytest.mark.asyncio
async def test_flush():
    temp_dir = tempfile.TemporaryDirectory()
    option = DbOption(temp_dir.name)
    option.immutable_chunk_num = 1
    option.major_threshold_with_sst_size = 3
    option.level_sst_magnification = 1
    option.max_sst_file_size = 1 * 1024
    db = TonboDB(option, User())
    # db = build_db()
    for i in range(0, 1000):
        if i % 100 == 0:
            await db.flush()
        await db.insert(
            User(
                id=i,
                age=i % 128,
                name=str(i * 10),
                email=str(i * 20),
                data=b"Hello Tonbo!",
            )
        )

    for i in range(0, 1000):
        user = await db.get(i)
        assert user == {
            "id": i,
            "name": str(i * 10),
            "email": str(i * 20),
            "age": i % 128,
            "data": b"Hello Tonbo!",
        }



from tonbo import Record, Column, DataType


@Record
class User:
    id = Column(DataType.Int64, name="id", primary_key=True)
    age = Column(datatype=DataType.UInt8, name="age", nullable=True)
    # nullable is `False` by default
    name = Column(DataType.String, name="name")
    height = Column(DataType.Int16, name="height")
    enabled = Column(DataType.Boolean, name="enabled")
    email = Column(DataType.String, name="email", nullable=True)
    data = Column(DataType.Bytes, name="data", nullable=True)




class CommitError(Exception):
    """Commit error"""
    pass

class DbError(Exception):
    """DB error"""
    pass

class DecodeError(Exception):
    """Decode error"""
    pass

class RecoverError(Exception):
    """Recover error"""
    pass

class ExceedsMaxLevelError(Exception):
    """Exceeds max level"""
    pass

class WriteConflictError(Exception):
    """Write conflict"""
    pass

class RepeatedCommitError(Exception):
    """Repeated commit the same transaction"""
    pass

class PathParseError(Exception):
    """Parse path error"""
    pass


from typing import final

@final
class AwsCredential:
    key_id: str
    secret_key: str
    token: str | None

    def __init__(
        self, key_id: str, secret_key: str, token: str | None = None
    ) -> None: ...

@final
class FsOptions:
    class Local: ...

    class S3:
        bucket: str
        credential: AwsCredential | None
        region: str | None
        sign_payload: bool | None
        checksum: bool | None

        def __init__(
            self,
            bucket: str,
            credential: AwsCredential | None = None,
            region: str | None = None,
            sign_payload: bool | None = None,
            checksum: bool | None = None,
        ) -> None: ...

def parse(path: str) -> str: ...

def from_filesystem_path(path: str) -> str:
    """Parse path to a filesystem relative path"""
    ...
def from_absolute_path(path: str) -> str:
    """Parse path to a filesystem absolute path"""
    ...
def from_url_path(path: str) -> str:
    """Parse an url path"""
    ...



import asyncio
import tempfile

from tonbo import Column, DataType, TonboDB, RecordBatch, DbOption, Record
from tonbo.fs import from_filesystem_path


@Record
class User:
    id = Column(DataType.UInt64, name="id", primary_key=True)
    age = Column(datatype=DataType.UInt8, name="age", nullable=True)
    name = Column(DataType.String, name="name")
    email = Column(DataType.String, name="email")


@Record
class Pet:
    id = Column(DataType.UInt64, name="id", primary_key=True)
    type = Column(DataType.String, name="type")
    name = Column(DataType.String, name="name")
    master = Column(datatype=DataType.UInt64, name="master", nullable=True)


@Record
class FoundRecord:
    user_id = Column(DataType.UInt64, name="user_id", primary_key=True)
    pet_id = Column(DataType.UInt64, name="pet_id")
    email = Column(DataType.String, name="email")
    status = Column(DataType.Boolean, name="status")


async def main():
    user_temp_dir = tempfile.TemporaryDirectory()
    pet_temp_dir = tempfile.TemporaryDirectory()
    found_temp_dir = tempfile.TemporaryDirectory()
    user_db = TonboDB(DbOption(from_filesystem_path(user_temp_dir.name)), User())
    course_db = TonboDB(DbOption(from_filesystem_path(pet_temp_dir.name)), Pet())
    found_db = TonboDB(DbOption(from_filesystem_path(found_temp_dir.name)), FoundRecord())

    user_txn = await user_db.transaction()
    for i in range(0, 100):
        user_txn.insert(User(id=i, age=i, name=str(i * 20), email=str(i * 100)))
    user_txn.commit()

    pet_txn = await course_db.transaction()
    pet_txn.insert(Pet(id=1, type="dog", name="Hachi", master=10))
    pet_txn.insert(Pet(id=2, type="cat", name="Kitty", master=21))
    pet_txn.insert(Pet(id=3, type="dragonfly", name="Tonbo", master=11))

    await pet_txn.commit()

    user_txn = await user_db.transaction()
    found_txn = await found_db.transaction()
    user_scan = await user_txn.scan(lower=None, high=None, projection=["id", "name", "email"])

    async for user in user_scan:
        pet_txn = await course_db.transaction()
        pet_scan = await pet_txn.scan(lower=None, high=None, projection=["master", "name"])
        async for pet in pet_scan:
            if user["id"] == pet["master"]:
                found_txn.insert(FoundRecord(user_id=user["id"], pet_id=pet["id"], email=user["email"], status=False))

    await found_txn.commit()

    found_txn = await found_db.transaction()
    found_scan = await found_txn.scan(lower=None, high=None)
    async for found in found_scan:
        print(found)


asyncio.run(main())



import asyncio
import random

import pytest_asyncio


def gen_string(max_size):
    size = gen_int(0, max_size)
    charset = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return ''.join(random.choices(charset, k=size))


def gen_bytes(max_size):
    return gen_string(max_size).encode("utf-8")


def gen_int(lower, high):
    return random.randint(lower, high)


# async support for pytest-benchmark
# https://github.com/ionelmc/pytest-benchmark/issues/66#issuecomment-1137005280
@pytest_asyncio.fixture
def aio_benchmark(benchmark, event_loop):
    def _wrapper(func, *args, **kwargs):
        if asyncio.iscoroutinefunction(func):

            @benchmark
            def _():
                return event_loop.run_until_complete(func(*args, **kwargs))
        else:
            benchmark(func, *args, **kwargs)

    return _wrapper



from tonbo import DbOption, Column, DataType, Record, TonboDB, Bound
import asyncio
import tempfile

@Record
class User:
    id = Column(DataType.Int64, name="id", primary_key=True)
    age = Column(DataType.Int16, name="age", nullable=True)
    name = Column(DataType.String, name="name", nullable=False)
    email = Column(DataType.String, name="email", nullable=True)
    data = Column(DataType.Bytes, name="data", nullable=True)


async def main():
    temp_dir = tempfile.TemporaryDirectory()

    db = TonboDB(DbOption(temp_dir.name), User())
    await db.insert(User(id=18, age=175, name="Alice"))

    record = await db.get(18)
    assert record == {
        "id": 18,
        "age": 175,
        "name": "Alice",
        "email": None,
        "data": None,
    }

    txn = await db.transaction()
    result = await txn.get(18)
    assert result == {
        "id": 18,
        "age": 175,
        "name": "Alice",
        "email": None,
        "data": None,
    }

    txn.insert(
        User(
            id=19,
            age=195,
            name="Bob",
            data=b"Hello Tonbo!",
            email="contact@tonbo.io",
        )
    )
    result = await txn.get(19)
    assert result == {
        "id": 19,
        "age": 195,
        "name": "Bob",
        "email": "contact@tonbo.io",
        "data": b"Hello Tonbo!",
    }

    await txn.commit()
    txn = await db.transaction()
    scan = await txn.scan(
        Bound.Excluded(18),
        None,
        limit=100,
        projection=["id", "email", "data"],
    )
    async for record in scan:
        assert record["age"] is None
        print(record)
    await txn.commit()


asyncio.run(main())



import asyncio
import os
import tempfile
import duckdb
import sqlite3
import pytest

from conftest import gen_int, gen_string, gen_bytes
from tonbo import Record, Column, DataType, TonboDB, DbOption
from tonbo.fs import from_filesystem_path

WRITE_TIME = 500000


@Record
class User:
    id = Column(DataType.Int64, name="id", primary_key=True)
    name = Column(DataType.String, name="name")
    email = Column(DataType.String, name="email", nullable=True)
    age = Column(DataType.UInt16, name="age")
    data = Column(DataType.Bytes, name="data")


async def duckdb_write(threads: int):
    con = duckdb.connect(config={"threads": threads})
    con.sql(
        "CREATE TABLE user (id INTEGER, name VARCHAR(20), email VARCHAR(20), age INTEGER, data VARCHAR(200))"
    )

    async def insert_task(start: int, end: int):
        txn = con.begin()
        for j in range(start, end):
            txn.execute(
                "INSERT INTO user VALUES (?, ?, ?, ?, ?)",
                [i, gen_string(20), gen_string(20), gen_int(0, 0xffff), gen_bytes(200)],
            )
        txn.commit()

    tasks = []
    for i in range(0, threads):
        tasks.append(insert_task(i * WRITE_TIME // threads, (i + 1) * WRITE_TIME // threads))

    await asyncio.gather(*tasks)
    con.commit()
    con.close()


async def tonbo_write(threads: int):
    temp_dir = tempfile.TemporaryDirectory()

    option = DbOption(from_filesystem_path(temp_dir.name))

    db = TonboDB(option, User())
    tasks = []

    async def insert_task(start: int, end: int):
        txn = await db.transaction()
        for j in range(start, end):
            txn.insert(
                User(
                    id=j,
                    age=gen_int(0, 0xffff),
                    name=gen_string(20),
                    email=gen_string(20),
                    data=gen_bytes(200),
                )
            )
        await txn.commit()

    for i in range(0, threads):
        tasks.append(insert_task(i * WRITE_TIME // threads, (i + 1) * WRITE_TIME // threads))

    await asyncio.gather(*tasks)
    await db.flush_wal()


async def tonbo_write_disable_wal(threads: int):
    temp_dir = tempfile.TemporaryDirectory()

    option = DbOption(from_filesystem_path(temp_dir.name))
    option.use_wal = False

    db = TonboDB(option, User())
    tasks = []

    async def insert_task(start: int, end: int):
        txn = await db.transaction()
        for j in range(start, end):
            txn.insert(
                User(
                    id=j,
                    age=gen_int(0, 0xffff),
                    name=gen_string(20),
                    email=gen_string(20),
                    data=gen_bytes(200),
                )
            )
        await txn.commit()

    for i in range(0, threads):
        tasks.append(insert_task(i * WRITE_TIME // threads, (i + 1) * WRITE_TIME // threads))

    await asyncio.gather(*tasks)


async def sqlite_write(threads: int):
    file = tempfile.NamedTemporaryFile()
    con = sqlite3.connect(file.name, check_same_thread=False)
    con.execute(
        "CREATE TABLE user (id INTEGER, name VARCHAR(20), email VARCHAR(20), age INTEGER, data VARCHAR(200))"
    )

    async def insert_task(start: int, end: int):
        for j in range(start, end):
            con.execute(
                "INSERT INTO user VALUES (?, ?, ?, ?, ?)",
                [j, gen_string(20), gen_string(20), gen_int(0, 0xffff), gen_bytes(200)],
            )
        con.commit()

    tasks = []
    for i in range(0, threads):
        tasks.append(insert_task(i * WRITE_TIME // threads, (i + 1) * WRITE_TIME // threads))
    await asyncio.gather(*tasks)
    con.commit()
    con.close()


@pytest.mark.parametrize("threads", [1])
@pytest.mark.benchmark(group="1 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_duckdb_one_task(aio_benchmark, threads):
    aio_benchmark(duckdb_write, threads)


@pytest.mark.parametrize("threads", [1])
@pytest.mark.benchmark(group="1 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_tonbo_one_task(aio_benchmark, threads):
    aio_benchmark(tonbo_write, threads)


@pytest.mark.parametrize("threads", [1])
@pytest.mark.benchmark(group="1 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_tonbo_disable_wal_one_task(aio_benchmark, threads):
    aio_benchmark(tonbo_write_disable_wal, threads)


@pytest.mark.parametrize("threads", [1])
@pytest.mark.benchmark(group="1 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_sqlite_one_task(aio_benchmark, threads):
    aio_benchmark(sqlite_write, threads)


@pytest.mark.parametrize("threads", [4])
@pytest.mark.benchmark(group="4 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_duckdb_four_task(aio_benchmark, threads):
    aio_benchmark(duckdb_write, threads)


@pytest.mark.parametrize("threads", [4])
@pytest.mark.benchmark(group="4 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_tonbo_four_task(aio_benchmark, threads):
    aio_benchmark(tonbo_write, threads)


@pytest.mark.parametrize("threads", [4])
@pytest.mark.benchmark(group="4 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_tonbo_disable_wal_four_task(aio_benchmark, threads):
    aio_benchmark(tonbo_write_disable_wal, threads)


@pytest.mark.parametrize("threads", [4])
@pytest.mark.benchmark(group="4 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_sqlite_four_task(aio_benchmark, threads):
    aio_benchmark(sqlite_write, threads)


@pytest.mark.parametrize("threads", [8])
@pytest.mark.benchmark(group="8 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_duckdb_eight_task(aio_benchmark, threads):
    aio_benchmark(duckdb_write, threads)


@pytest.mark.parametrize("threads", [8])
@pytest.mark.benchmark(group="8 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_tonbo_eight_task(aio_benchmark, threads):
    aio_benchmark(tonbo_write, threads)


@pytest.mark.parametrize("threads", [8])
@pytest.mark.benchmark(group="8 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_tonbo_disable_wal_eight_task(aio_benchmark, threads):
    aio_benchmark(tonbo_write_disable_wal, threads)


@pytest.mark.parametrize("threads", [8])
@pytest.mark.benchmark(group="8 async task")
@pytest.mark.skipif("BENCH" not in os.environ, reason="benchmark")
def test_sqlite_eight_task(aio_benchmark, threads):
    aio_benchmark(sqlite_write, threads)



import asyncio
import os

from tonbo import DbOption, Column, DataType, Record, TonboDB
from tonbo.fs import from_filesystem_path, FsOptions


@Record
class User:
    id = Column(DataType.Int64, name="id", primary_key=True)
    name = Column(DataType.String, name="name")
    email = Column(DataType.String, name="email", nullable=True)
    age = Column(DataType.UInt8, name="age")
    data = Column(DataType.Bytes, name="data")


async def main():
    if not os.path.exists("db_path/user/l0"):
        os.makedirs("db_path/user/l0")
    if not os.path.exists("db_path/user/l1"):
        os.makedirs("db_path/user/l1")

    option = DbOption(from_filesystem_path("db_path/user"))
    option.level_path(0, from_filesystem_path("db_path/user/l0"), FsOptions.Local())
    option.level_path(1, from_filesystem_path("db_path/user/l1"), FsOptions.Local())

    option.immutable_chunk_num = 1
    option.major_threshold_with_sst_size = 3
    option.level_sst_magnification = 1
    option.max_sst_file_size = 1 * 1024

    db = TonboDB(option, User())
    for i in range(0, 1000):
        if i % 50 == 0:
            await db.flush()
        await db.insert(
            User(
                id=i,
                age=i % 128,
                name=str(i * 10),
                email=str(i * 20),
                data=b"Hello Tonbo!",
            )
        )

    for i in range(0, 1000):
        user = await db.get(i)
        assert user == {
            "id": i,
            "name": str(i * 10),
            "email": str(i * 20),
            "age": i % 128,
            "data": b"Hello Tonbo!",
        }


asyncio.run(main())



import asyncio
import tempfile

from tonbo import Column, DataType, TonboDB, RecordBatch, DbOption, Record
from tonbo.fs import from_filesystem_path


@Record
class User:
    id = Column(DataType.Int64, name="id", primary_key=True)
    age = Column(datatype=DataType.UInt8, name="age", nullable=True)
    name = Column(DataType.String, name="name")
    email = Column(DataType.String, name="email", nullable=True)
    data = Column(DataType.Bytes, name="data", nullable=True)


async def main():
    temp_dir = tempfile.TemporaryDirectory()
    db = TonboDB(DbOption(from_filesystem_path(temp_dir.name)), User())
    batch = RecordBatch()
    for i in range(0, 100):
        # you must add record to `RecordBatch` one by one
        batch.append(User(id= i, age=i, name=str(i * 20)))

    await db.insert_batch(batch)

    for i in range(0, 100):
        user = await db.get(i)
        assert user == { "id": i, "age": i, "name": str(i * 20), "email": None, "data": None }


asyncio.run(main())



from typing import Any, AsyncIterable, final
from enum import Enum, auto
from tonbo import error as error
from tonbo.fs import FsOptions

@final
class Record: ...

@final
class Bound(Enum):
    """Tonbo range for scan. None for unbounded"""

    Included = auto()
    Excluded = auto()

@final
class DataType(Enum):
    """Tonbo data type."""

    UInt8 = auto()
    UInt16 = auto()
    UInt32 = auto()
    UInt64 = auto()
    Int8 = auto()
    Int16 = auto()
    Int32 = auto()
    Int64 = auto()
    String = auto()
    Boolean = auto()
    Bytes = auto()

@final
class Column:
    """Column represents properties of a field."""

    name: str
    datatype: DataType
    nullable: bool
    primary_key: bool
    def __init__(
        self,
        datatype: DataType,
        name: str,
        nullable: bool = False,
        primary_key: bool = False,
    ) -> None: ...

@final
class RecordBatch:
    """Column represents properties of a field."""
    def __init__(self) -> None: ...
    def append(self, record: object) -> None: ...

class DbOption:
    """Tonbo configurations."""

    clean_channel_buffer: int
    immutable_chunk_num: int
    level_sst_magnification: int
    major_default_oldest_table_num: int
    major_threshold_with_sst_size: int
    max_sst_file_size: int
    version_log_snapshot_threshold: int
    use_wal: bool
    wal_buffer_size: int
    path: str

    def __init__(self, path: str) -> None:
        """Create a new :py:class:`DbOption` with the given path. Note: the path must exist"""
        ...

    def level_path(self, level: int, path: str, fs_options: FsOptions) -> None:
        """Set path for assigned level

        Args:
            level: Level for output.
            path: Path for output
            fs_options: Local or S3
        """
        ...

@final
class Transaction:
    """Tonbo transaction."""

    async def get(
        self, key: Any, projection: list[str] = ["*"]
    ) -> dict[str, Any] | None:
        """Get record from db.

        Args:
            key: Primary key of record.
            projection: fields to projection
        """
        ...
    def insert(self, record: object) -> None:
        """Insert record to db."""
        ...
    def remove(self, key: Any) -> None:
        """Remove record from db.
        Args:
            key: Primary key of record.
        """
        ...
    async def scan(
        self,
        lower: Bound | None,
        high: Bound | None,
        limit: int | None,
        projection: list[str] = ["*"],
    ) -> AsyncIterable[dict[str, Any]]:
        """Create an async stream for scanning.

        Args:
            lower: Lower bound of range. Use None represent unbounded.
            high: High bound of range. Use None represent unbounded.
            limit: max number records to scan
            projection: fields to projection
        """
        ...
    async def commit(self) -> None:
        """Commit :py:class:`Transaction`."""
        ...

@final
class TonboDB:
    def __init__(self, option: DbOption, schema: Any) -> None:
        """Create a new :py:class:`TonboDB` with the given configuration options.

        Args:
            option: Configuration options.
            schema: Schema of record.
        """
        ...
    async def get(self, key: Any) -> dict[str, Any] | None:
        """Get record from db.

        Args:
            key: Primary key of record.
        """
        ...
    async def insert(self, record: object) -> None:
        """Insert a record to db."""
        ...
    async def insert_batch(self, record_batch: RecordBatch) -> None:
        """Insert :py:class:`RecordBatch` to db."""
        ...
    async def remove(self, key: Any) -> None:
        """Remove record from db.

        Args:
            key: Primary key of record.
        """
        ...
    async def transaction(self) -> Transaction:
        """Create a new :py:class:`Transaction`."""
        ...
    async def flush(self) -> None:
        """Try to execute compaction."""
        ...
    async def flush_wal(self) -> None:
        """Flush wal manually."""
        ...
